import wrapt

from django_aws_xray.traces import trace_sql


def key(db, attr):
    return 'db.%s.%s.%s' % (db.client.executable_name, db.alias, attr)


def _get_query_type(query):
    return (query.split(None, 1) or ['__empty__'])[0].lower()


def patched_execute(func, instance, args, kwargs):
    query = args[0]
    name = key(instance.db, 'execute.%s' % _get_query_type(query))
    with trace_sql(name, instance.db, query):
        return func(*args, **kwargs)


def patched_executemany(func, instance, args, kwargs):
    query = args[0]
    name = key(instance.db, 'executemany.%s' % _get_query_type(query))
    with trace_sql(name, instance.db, query):
        return func(*args, **kwargs)


def patched_callproc(func, instance, args, kwargs):
    query = args[0]
    name = key(instance.db, 'callproc.%s' % _get_query_type(query))
    with trace_sql(name, instance.db, query):
        return func(*args, **kwargs)


def patch():
    wrapt.wrap_function_wrapper(
        'django.db.backends.utils', 'CursorWrapper.execute', patched_execute)
    wrapt.wrap_function_wrapper(
        'django.db.backends.utils', 'CursorWrapper.executemany', patched_executemany)
    wrapt.wrap_function_wrapper(
        'django.db.backends.utils', 'CursorWrapper.callproc', patched_callproc)



from functools import partial, wraps

from django_aws_xray import traces


def patch_method(target, name, external_decorator=None):

    def decorator(patch_function):
        original_function = getattr(target, name)

        @wraps(patch_function)
        def wrapper(*args, **kw):
            return patch_function(original_function, *args, **kw)

        setattr(target, name, wrapper)
        return wrapper

    return decorator


def wrapped(method, key, *args, **kw):
    with traces.trace(key):
        return method(*args, **kw)


def wrap(method, key, *args, **kw):
    return partial(wrapped, method, key, *args, **kw)



import threading
import time
import uuid
from contextlib import contextmanager

from django_aws_xray import records
from django_aws_xray.connection import Connection

tls = threading.local()
tls.trace_id = None
tls.current_trace = None


def set_current_trace(trace):
    tls.current_trace = trace


def get_current_trace():
    return getattr(tls, 'current_trace', None)


class Trace:

    def __init__(self, root, parent=None, sampled=None):
        self.root = root
        self.parent = parent
        self.sampled = sampled
        self._connection = Connection()
        self._segment_stack = list()
        self._segment_buffer = list()

    def send(self):
        if self.sampled:
            for record in self._segment_buffer:
                self._connection.send(record)
            self._segment_buffer.clear()

    @property
    def http_header(self):
        parts = [
            ('Root', self.root),
            ('Parent', self.parent),
            ('Sampled', '1' if self.sampled else None),
        ]
        return ';'.join('%s=%s' % (k, v) for k, v in parts if v is not None)

    @contextmanager
    def track(self, name):
        # Create the correct record type
        if len(self._segment_stack) == 0:
            record = records.SegmentRecord(
                name=name,
                trace_id=self.root)
        else:
            parent = self._segment_stack[-1]
            record = records.SubSegmentRecord(
                name=name,
                trace_id=self.root,
                parent_id=parent.id)

        # Push the record on our stack
        self._segment_stack.append(record)

        record.trace_id = self.root
        record.start_time = time.time()
        try:
            yield record
        finally:
            assert self._segment_stack.pop() is record
            record.end_time = time.time()
            self._segment_buffer.append(record)

    @classmethod
    def generate_new(cls, sampled=True):
        root = '1-%08x-%s' % (int(time.time()), uuid.uuid4().hex[:24])
        return cls(root=root, parent=None, sampled=sampled)

    @classmethod
    def from_http_header(cls, header_value, sampled):
        """

        Root=1-5759e988-bd862e3fe1be46a994272793;Parent=53995c3f42cd8ad8;Sampled=1

        """
        parts = header_value.lower().split(';')
        data = {
            'sampled': '1' if sampled else '0'
        }

        for part in parts:
            subparts = part.split('=', 1)
            if len(subparts) == 2:
                data[subparts[0]] = subparts[1]

        root = data.get('root')
        parent = data.get('parent')
        sampled = data.get('sampled') == '1'
        return cls(root=root, parent=parent, sampled=sampled)



import wrapt

from django_aws_xray.traces import trace_http


def patched_request(func, instance, args, kwargs):
    method = kwargs.get('method') or args[0]
    url = kwargs.get('url') or args[1]
    name = 'requests.request'

    with trace_http(name, method, url) as trace:
        response = func(*args, **kwargs)
        trace.response_status_code = response.status_code
        return response


def patch():
    """ Monkeypatch the requests library to trace http calls. """
    wrapt.wrap_function_wrapper('requests', 'Session.request', patched_request)



from contextlib import contextmanager

from django_aws_xray import records, xray

MAX_SQL_QUERY_LENGTH = 8192


@contextmanager
def trace(name):
    trace = xray.get_current_trace()
    if trace:
        with trace.track(name):
            yield
    else:
        yield


@contextmanager
def trace_sql(name, db, query):
    sql_record = records.SqlRecord(
        sanitized_query=query.strip()[:MAX_SQL_QUERY_LENGTH],
        database_type=db.vendor)

    trace = xray.get_current_trace()
    if trace:
        with trace.track(name) as record:
            record.sql = sql_record
            yield sql_record
    else:
        yield sql_record


@contextmanager
def trace_http(name, method, url):
    http_record = records.HttpRecord(
        request_method=method,
        request_url=url)

    trace = xray.get_current_trace()
    if trace:
        with trace.track(name) as record:
            record.http = http_record
            yield http_record
    else:
        yield http_record



import wrapt

from django_aws_xray.traces import trace_http


def patched_execute_command(func, instance, args, kwargs):

    # TODO: Metadata / Annotations
    # instance.connection_pool.connection_kwargs['host']
    # instance.connection_pool.connection_kwargs['db']

    command = args[0]
    key = None
    if command in ('GET', 'SET', 'DELETE', 'INCR', 'DECR'):
        key = str(args[1])

    with trace_http('redis', command, key) as trace:
        response = func(*args, **kwargs)
        trace.response_status_code = response.status_code
        return response


def patch():
    """ Monkeypatch the requests library to trace http calls. """
    try:
        wrapt.wrap_function_wrapper(
            'redis.client', 'Redis.execute_command', patched_execute_command)
        wrapt.wrap_function_wrapper(
            'redis.client', 'StrictRedis.execute_command', patched_execute_command)
    except ModuleNotFoundError:
        pass
